import os
import glob
import numpy as np
import pandas as pd
from projectframe import ProjectFrame


def write_h5(pf: ProjectFrame, h5_filename: str) -> None:
    """
    Write project frame to HDF5 file

    Parameters
    ----------
        pf : :class: `project_frame.ProjectFrame`
            Project frame
        h5_filename : str
            Name of HDF5 file to write to

    Returns
    -------
        None
    """

    # Write project frame to HDF5 file
    pf.write_h5(h5_filename)

    return None



from .write_text import *
from .write_h5 import *

__doc__="""\

IO API
============
.. currentmodule:: cfdna

Generic
-------
.. autosummary::
   :toctree: .

   write_seg
   write_anno
   write_anno_seg
   write_h5

"""


import numpy as np
import pandas as pd
from ngsfragments import Fragments
from intervalframe import IntervalFrame
import glob

# Local imports
from ...core import cfDNA
from .wps import wps
from ...io.read.read_bam import readBam


def calculate_relative_nhits(frags: Fragments,
                             start: int,
                             end: int,
                             chrom: str,
                             min_length: int = 120,
                             max_length: int = 220):
    """
    """

    # Calculate gene score
    gene_score = frags.frags.nhits(start,
                                    end,
                                    chrom,
                                    min_length,
                                    max_length)
    
    # Calculate upstream score
    up_start = max(start - 10000, 0)
    up_end = start
    upstream_score = frags.frags.nhits(up_start,
                                        up_end,
                                        chrom,
                                        min_length,
                                        max_length)
    # Calculate downstream score
    down_start = end
    down_end = end + 10000
    downstream_score = frags.frags.nhits(down_start,
                                        down_end,
                                        chrom,
                                        min_length,
                                        max_length)
    
    # Calculate score
    gene_score = gene_score / (end - start)
    gene_score = max(gene_score, 0.000001)
    flank_score = (upstream_score + downstream_score) / ((up_end - up_start) + (down_end - down_start))
    flank_score = max(flank_score, 0.000001)
    score = np.log2(gene_score / flank_score)

    return score


def calculate_relative_wps(wps: pd.Series,
                             start: int,
                             end: int):
    """
    """

    # Calculate gene score
    gene_score = wps.loc[start:end].values.sum()

    # Calculate upstream score
    up_start = max(start - 10000, wps.index.values[0])
    up_end = start
    upstream_score = wps.loc[up_start:up_end].values.sum()

    # Calculate downstream score
    down_start = end
    down_end = min(end + 10000, wps.index.values[-1])
    downstream_score = wps.loc[down_start:down_end].values.sum()

    # Calculate score
    gene_score = gene_score / (end - start)
    if gene_score == 0:
        gene_score = 0.001
    flank_score = (upstream_score + downstream_score) / ((up_end - up_start) + (down_end - down_start))
    if flank_score == 0:
        flank_score = 0.001
    score = np.log2((gene_score+1) / (flank_score+1))

    return score


def gene_relative_enrichment(frags: Fragments,
                             genome: str = "hg19",
                             min_length: int = 120,
                             max_length: int = 220,
                             use_wps: bool = False,
                             protection: int = 120):
    """
    """
    # Get windows
    if genome == "hg19":
        import hg19genome as Genome
    else:
        raise NotImplementedError("Only hg19 is supported currently")

    # Get genes
    #genes = Genome.get_gene_body(upstream=5000, gene_type="protein_coding")
    genes = Genome.get_tss(upstream=1000, downstream=1000)

    # Determine chromosomes
    chromosomes = frags.chroms

    # Iterate over chromosomes
    scores = pd.Series(np.zeros(genes.shape[0]),
                       index = genes.loc[:,"Gene"].values)
    j = 0
    for chrom in chromosomes:
        print(chrom)
        # Get genes on chromosome
        chrom_genes = genes.loc[chrom,:]
        
        # Calculate WPS
        if use_wps:
            wps_scores = wps(frags, chrom, protection, min_length, max_length)

        for i in range(chrom_genes.shape[0]):
            if use_wps:
                score = calculate_relative_wps(wps_scores[chrom],
                                                chrom_genes.index[i].start,
                                                chrom_genes.index[i].end)
            else:
                score = calculate_relative_nhits(frags,
                                                chrom_genes.index[i].start,
                                                chrom_genes.index[i].end,
                                                str(chrom),
                                                min_length,
                                                max_length)
            scores[j] = score
            j += 1

    return scores


def multi_rel_enrich(directory: str,
                        genome: str = "hg19",
                        min_length: int = 121,
                        max_length: int = 375,
                        nthreads: int = 1,
                        use_wps: bool = False):
    """
    """

    files = glob.glob(directory+"*.bam")

    print(files[0])
    frags = readBam(files[0], nthreads=nthreads)
    scores = gene_relative_enrichment(frags, genome=genome, min_length=min_length, max_length=max_length, use_wps=use_wps)
    scores = scores.to_frame().T
    scores.index = [files[0]]
    for bam in files[1:]:
        print(bam)
        frags = readBam(bam, nthreads=nthreads)
        scores.loc[bam,:] = gene_relative_enrichment(frags, genome=genome, min_length=min_length, max_length=max_length, use_wps=use_wps)

    return scores


import ngsfragments as ngs
from ngsfragments import Fragments
from ..core.core import cfDNA


def gene_activity(frags: Fragments,
                  cfdna_object: cfDNA,
                  genome_version: str = "hg19",
                  feature: str = "gene",
                  min_length: int = 120,
                  max_length: int = 220,
                  verbose: bool = False) -> cfDNA:
    """
    """

    # Get gene activity
    gene_activity = ngs.metrics.gene_activity(frags,
                                            genome_version=genome_version,
                                            feature=feature,
                                            min_length=min_length,
                                            max_length=max_length,
                                            verbose=verbose)
    gene_activity = ngs.metrics.correct_gene_activity(frags,
                                                     gene_activity,
                                                     correct_cnv = False,
                                                     genome_version = genome_version,
                                                     feature = feature,
                                                     verbose = verbose)

    cfdna_object.add_values("gene_activity", gene_activity)

    return cfdna_object


import ngsfragments as ngs
import numpy as np
import pandas as pd
from ailist import LabeledIntervalArray
import scipy
from scipy.signal import lfilter
from scipy.stats import trim_mean
from scipy.signal import periodogram, welch
from scipy.interpolate import Rbf

# Local import
from ...core import cfDNA


def wps(frags, chrom=None, protection=120, min_length=120, max_length=210, normalize=True, norm_method="mean"):
    """
    Calculate Window Protection Score (WPS)

    Parameters
    ----------
        frags : ngsfragments.fragments
            Fragments object
        chroms : str or list
            Name of chromosome
        protection : int
            Protection window
        min_length : int
            Minimum DNA fragment length
        max_length : int
            Maximum fragment length
        normalize : bool
            Whether to normalize
        norm_method : str
            Normalization method

    Returns
    -------
        wps_score : pandas.Series
            Window protection scores
    """
    
    # Caclulate WPS
    wps_score = frags.wps(chrom=chrom, protection=protection, min_length=min_length, max_length=max_length)
    if normalize:
        ngs.utilities.normalize_wps(wps_score, method=norm_method)

    return wps_score


def coverage(frags, chrom=None, min_length=120, max_length=210, normalize=True, norm_method="mean"):
    """
    Calculate Window Protection Score (WPS)

    Parameters
    ----------
        frags : ngsfragments.fragments
            Fragments object
        chroms : str or list
            Name of chromosome
        min_length : int
            Minimum DNA fragment length
        max_length : int
            Maximum fragment length
        normalize : bool
            Whether to normalize
        norm_method : str
            Normalization method

    Returns
    -------
        cov : pandas.Series
            Coverage
    """
    
    # Caclulate coverage
    cov = frags.coverage(chrom=chrom, min_length=min_length, max_length=max_length)
    if normalize:
        ngs.utilities.normalize_wps(cov, method=norm_method)

    return cov


def predict_nucleosomes(fragments,
                        protection = 120,
                        merge_distance = 5,
				        min_length = 120,
                        max_length = 500,
                        verbose = False):
    """
    Call peaks from WPS

    Parameters
    ----------
        chrom
            str (chromosomes to call for)

    Returns
    -------
        peaks
            dict[chrom:AIList] (WPS peaks)
    """

    # Iterate over chromosomes
    chroms = fragments.frags.unique_labels

    # Initialize peaks
    total_peaks = LabeledIntervalArray()

    # Iterate over chromosomes
    for chrom in chroms:
        if verbose: print(chrom, flush=True)
        wps = fragments.frags.wps(protection, chrom, min_length, max_length)
        wps[chrom].values[:] = ngs.peak_calling.CallPeaks.normalize_signal(wps[chrom].values)

        peaks = ngs.peak_calling.CallPeaks.call_peaks(wps[chrom], str(chrom), merge_distance, 50, 190)
        if peaks.size < 10:
            continue
        midpoints = (peaks.starts + ((peaks.ends - peaks.starts) / 2)).astype(int)
        standard_peaks = LabeledIntervalArray()
        standard_peaks.add(midpoints - 84, midpoints + 84, np.repeat(chrom, len(midpoints)))
        standard_peaks = standard_peaks.merge()
        midpoints = (standard_peaks.starts + ((standard_peaks.ends - standard_peaks.starts) / 2)).astype(int)
        standard_peaks = LabeledIntervalArray()
        standard_peaks.add(midpoints - 84, midpoints + 84, np.repeat(chrom, len(midpoints)))
        total_peaks.append(standard_peaks)

    return total_peaks


def merge_nucleosomes(peaks1, peaks2, merge_distance=10):
    """
    Merge two sets of nucleosomes

    Parameters
    ----------
        peaks1
            dict[chrom:AIList]
        peaks2
            dict[chrom:AIList]
        merge_distance
            int

    Returns
    -------
        merged_peaks
            dict[chrom:AIList]
    """

    # Merge peaks
    merged_peaks = peaks1.union(peaks2).merge(merge_distance)
    
    # Re-center peaks
    midpoints = (merged_peaks.starts + ((merged_peaks.ends - merged_peaks.starts) / 2)).astype(int)
    standard_peaks = LabeledIntervalArray()
    standard_peaks.add(midpoints - 84, midpoints + 84, merged_peaks.labels)

    return standard_peaks


def internucleosomal_distances(peaks, max_distance=1000):
    """
    Determine distance between nucleosomes
    """

    # Iterate over chromosomes
    chroms = peaks.unique_labels

    # Initialize distances
    distances = []

    # Iterate over chromosomes
    for chrom in chroms:
        chrom_peaks = peaks.get(chrom)
        midpoints = (chrom_peaks.starts + ((chrom_peaks.ends - chrom_peaks.starts) / 2)).astype(int)

        pdist = np.diff(midpoints)
        distances.extend(pdist[pdist < max_distance])

    distances = np.array(distances)

    return distances


def peak_distances(self, fragments, peaks=None, bin_size=100000, max_distance=10000):
    """
    Determine distance between peaks

    Parameters
    ----------
        peaks
            dict[chrom:AIList]
        bin_size
            int
        max_distance
            int
        smooth
            bool

    Returns
    -------
        peak_bins
            dict[pandas.Series]
    """

    # Calculate peaks
    if peaks is None:
        peaks = self.predict_nucleosomes(fragments)

    # Create p_dist IntervalFrame
    starts = np.array([],dtype=int)
    chroms = np.array([], dtype="U25")
    for chrom in peaks:
        new_starts = np.arange(0, self.chrom_sizes[chrom]+bin_size, bin_size)
        starts = np.append(starts, new_starts)
        chroms = np.append(chroms, np.repeat(chrom, len(new_starts)))
    ends = starts + bin_size
    p_dist = IntervalFrame.from_array(starts, ends, labels=chroms)

    # Iterate over intervals
    p_dist.df.loc[:,"mean_dist"] = 0
    for i, interval in enumerate(p_dist.index):
        chrom = interval.label
        try:
            peaks[chrom]
            overlaps = peaks[chrom].intersect(interval.start, interval.end, label=chrom)
            if len(overlaps) > 3:
                p_dist.df.loc[i,"mean_dist"] = np.mean(overlaps.extract_ends()[:-1] - overlaps.extract_starts()[1:])
        except KeyError:
            pass

    # Remove those greater than max_distance
    p_dist = p_dist.iloc[p_dist.df.loc[:,"mean_dist"].values < max_distance, :]

    return p_dist


def spec_pgram(x, smoothed=False):
    """
    """
    
    # Compute the periodogram
    frequencies, power_spectrum = periodogram(x,
                                              window="hann",
                                              nfft=None,
                                              return_onesided=True,
                                              scaling="density",
                                              detrend="linear")
    if smoothed:
        # Smooth the periodogram using Welch's method
        frequencies, power_spectrum = welch(x,
                                          window="hann",
                                          nperseg=None,
                                          noverlap=None,
                                          detrend="linear",
                                          scaling="density",
                                          return_onesided=True)
    return frequencies, power_spectrum


def wps_gene_fft(fragments,
                 genome_version = "hg38",
                 protection = 120,
				 min_length = 120,
                 max_length = 1000,
                 freq_range = (120, 280),
                 scale = True,
                 feature = "gene",
                 verbose = False):
    """
    """

    import genome_info

    genome = genome_info.GenomeInfo(genome_version)
    #genes = genome.get_intervals("tss", downstream=10000, filter_column="gene_type", filter_selection="protein_coding")
    if feature == "gene":
        genes = genome.get_intervals("gene", filter_column="gene_type", filter_selection="protein_coding")
    elif feature == "tss":
        genes = genome.get_intervals("tss", upstream=5000, downstream=5000, filter_column="gene_type", filter_selection="protein_coding")
    elif feature == "gene_body":
        genes = genome.get_intervals("gene", upstream=5000, filter_column="gene_type", filter_selection="protein_coding")

    # Iterate over chromosomes
    chroms = fragments.frags.unique_labels

    # Initialize scores
    scores = pd.DataFrame(np.zeros((genes.shape[0], freq_range[1]-freq_range[0])), index=genes.loc[:,"gene_name"].values)
    #scores = {}
    b = np.array([1])
    a = np.append(1, -(1 / np.arange(5, 100, 4)))

    # Iterate over chromosomes
    k = 0
    for chrom in chroms:
        if verbose: print(chrom, flush=True)
        wps = fragments.frags.wps(protection, chrom, min_length, max_length)
        wps[chrom].values[:] = ngs.peak_calling.CallPeaks.normalize_signal(wps[chrom].values)

        chrom_genes = genes.loc[chrom,:]

        for i in range(chrom_genes.shape[0]):
            interval = chrom_genes.index[i]
            tdataA1 = wps[chrom].loc[interval.start:interval.end].values
            x = np.append(tdataA1[:300], tdataA1)
            tdataA1 = lfilter(b, a, x)[300:]
            tdataA1 = tdataA1 - trim_mean(tdataA1, 0.1)
            resA1 = spec_pgram(tdataA1, smoothed=False)
            freq = 1 / resA1[0][1:]
            chosen = np.logical_and(freq >= freq_range[0], freq <= freq_range[1])
            if chosen.sum() < 5:
                continue
            freq = np.round(freq[chosen])
            fft_values = pd.Series(resA1[1][1:][chosen], index=freq)
            fft_values = fft_values.groupby(by=fft_values.index.values.astype(int)).mean()
            #print(fft_values, fft_values.shape, chosen, chosen.sum(), flush=True)
            interp_func = Rbf(fft_values.index, fft_values.values, smooth=0.1)
            fft_values = pd.Series(interp_func(np.arange(freq_range[0]-5, freq_range[1]+5)), index = np.arange(freq_range[0]-5, freq_range[1]+5))
            fft_values = fft_values.rolling(window=5, center=True).mean()
            fft_values = fft_values.loc[np.arange(freq_range[0], freq_range[1])]
            if scale:
                fft_values = (fft_values - fft_values.min()) / (fft_values.max() - fft_values.min())
            scores.loc[chrom_genes.df.loc[:,"gene_name"].values[i],:] = fft_values.values

    # Edit range
    scores.columns = np.arange(freq_range[0], freq_range[1])

    return scores


def wps_windows(cfdna_object: cfDNA,
                frags: ngs.Fragments,
                key: str = "tss_wps",
                protection: int = 120,
                min_length: int = 120,
                max_length: int = 210,
                feature: str = "tss",
                smooth: bool = False) -> cfDNA:
    """
    Calculate Window Protection Score (WPS) for windows

    Parameters
    ----------
        cfdna_object : cfDNA
            cfDNA object
        frags : ngs.Fragments
            Fragments object
        key : str
            Key for storing the WPS (default: "tss_wps")
        protection : int
            Protection window (default: 120)
        min_length : int
            Minimum DNA fragment length (default: 120)
        max_length : int
            Maximum fragment length (default: 210)
        feature : str
            Feature (default: "tss")
        smooth : bool
            Whether to smooth (default: False)
        
    Returns
    -------
        cfdna_object : cfDNA
            cfDNA object

    
    """

    scores = ngs.metrics.wps_windows(frags,
                                    protection = protection,
                                    min_length = min_length,
                                    max_length = max_length,
                                    feature = feature,
                                    smooth = smooth)
    
    cfdna_object.add_obs_values(list(cfdna_object.obs)[0], key, scores)

    return cfdna_object


from typing import Dict, List, Tuple
import pandas as pd
import numpy as np
import os
from multiprocessing import Pool
from joblib import Parallel, delayed
from ...core.core import cfDNA
from ...utilities.h5_utilities import merge_h5
from ..fragmentation.frag_pattern import fragment_profile
from ..cnv.segmentation import call_cnv_pipline


def call_gene_wps():
    pass

def call_tss_wps():
    pass

def call_frag_profile(cfdna_object, frags, bin_size=100000, bin_bias_h5_fn=None, bias_correct=True, smooth=True, verbose=False):
    fragment_profile(cfdna_object, frags, bin_size=100000, bin_bias_h5_fn=None, bias_correct=True, smooth=True, verbose=False)

def call_cnv_pipeline(cfdna_object, frags, cnv_binsize=cnv_binsize, hmm_binsize=hmm_binsize, method=method, outlier_smooth=outlier_smooth,
                         gauss_smooth=gauss_smooth, bcp_cutoff=bcp_cutoff, normal=normal, ploidy=ploidy, estimatePloidy=estimatePloidy,
                         minSegmentBins=minSegmentBins, maxCN=maxCN, verbose=verbose):
    call_cnv_pipline(cfdna_object, frags, cnv_binsize=cnv_binsize, hmm_binsize=hmm_binsize, method=method, outlier_smooth=outlier_smooth,
                         gauss_smooth=gauss_smooth, bcp_cutoff=bcp_cutoff, normal=normal, ploidy=ploidy, estimatePloidy=estimatePloidy,
                         minSegmentBins=minSegmentBins, maxCN=maxCN, verbose=verbose)


def pipeline(bam_filename: str,
             functions: List[str],
             **kwargs: int) -> str:
    """
    Pipe functions together while optimizing for reducing
    redundent calculations

    Parameters
    ----------
        bam_filename : str
            Bam file name
        functions : List[str]
            List of functions
        **kwargs : Iterable
            Key word arguements

    Returns
    -------
        h5_file : str
            Name of output h5 file
    """

    # Function calls
    function_calls = {"gene_wps": (call_gene_wps, "wps"),
                      "tss_wps": (call_tss_wps, "wps"),
                      "fragment_profile": (call_frag_profile, "frag"),
                      "cnv": (call_cnv_pipeline, "cnv")}
    
    # Determine functions
    func_typing = {"wps": [],
                   "frag": [],
                   "cnv": []}
    for func in functions:
        func_typing[function_calls[func][1]].append(function_calls[func][0])

    # Initialize cfdna object
    cfdna_object = cfDNA()

    # Read bam file
    frags = readBam(bam_filename)

    # Iterate over wps functions
    for func in func_typing["wps"]:
        analysis_func = func_typing["wps"][func]
        
        # Iterate over chromosomes
        for chrom in frags.frags.unique_labels:
            wps_scores = wps(frags, chrom=chrom)
            analysis_func(cfdna_object, wps_scores)


def multi_apply(bam_files, function, njobs, output_h5=None, merge=True, cleanup=True, **kwargs):
    """
    """

    function_calls = {"gene_wps": call_gene_wps,
                      "tss_wps": call_tss_wps,
                      "fragment_profile": call_frag_profile,
                      "cnv": call_cnv_pipeline}

    # Call function and write h5 files
    h5_filenames = Parallel(n_jobs=njobs)(delayed(function)(bam, **kwargs) for bam in bam_files)

    # Merge
    if merge:
        if output_h5 is None:
            output_h5 = function.__name__ + "_appied.h5"
        merge_h5(h5_filenames, output_h5, progress_bar=False)

    # Clean up
    if merge and cleanup:
        for filename in h5_filenames:
            os.remove(filename)

    return None


import os
import glob
import numpy as np
import pandas as pd
import h5py
from ..core.core import cfDNA
from ..io.read.read_h5 import read_h5
from ..io.write.write_h5 import write_h5


def printProgressBar(iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█', print_end = "\r"):
    """
    Call in a loop to create terminal progress bar
    
    Parameters
    ----------
        iteration
            int (current iteration)
        total
            int (total iterations)
        prefix
            str (prefix string [defualt:''])
        suffix
            str (suffix string [defualt:''])
        decimals
            int (positive number of decimals in percent complete [defualt:1])
        length
            int (character length of bar [defualt:100])
        fill
            str (bar fill character [defualt:'█'])
        print_end
            str (end character [defualt:'\r'])

    Returns
    -------
        None
    """
    
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end = print_end)
    
    # Print New Line on Complete
    if iteration == total: 
        print()


def find_h5_filenames(names_list, directory="./"):
    """
    Search directory for h5 file names

    Parameters
    ----------
        names_list
            list
        directory
            str

    Returns
    -------
        h5_filenames
            list
    """

    # Record files in directory
    directory_files = glob.glob(os.path.join(directory, "*.h5"), recursive=True)

    # Iterate over names in list
    h5_filenames = []
    for name in names_list:
        found = False
        for dir_fn in directory_files:
            if name in dir_fn:
                h5_filenames.append(dir_fn)
                found = True
                break
        # Could not find file
        if found == False:
            raise FileNotFoundError("Could not find file for " + name)
    
    return h5_filenames


def update_h5(reference_h5, query_h5):
    """
    Go through and update h5 in place
    """

    # Open files
    ref = h5py.File(reference_h5, "r+")
    query = h5py.File(query_h5, "r")

    pass


def copy_h5(h5_filename, output_h5):
    """
    """

    # Create new h5 file to copy to
    #out_f = h5py.File(output_h5, "w")
    #original_f = h5py.File(h5_filename)

    # Read data
    orig_cfdna = cfDNA()
    read_h5(orig_cfdna, h5_filename)

    # Write data
    write_h5(orig_cfdna, output_h5)


def merge_cfdna_objects(cfdna_object1, cfdna_object2):
    """
    """

    # Merge anno
    cfdna_object1._anno = pd.concat([cfdna_object1._anno, cfdna_object2._anno]).drop_duplicates()

    # Merge intervals
    for key in cfdna_object2.intervals:
        cfdna_object1.add_intervals(key, cfdna_object2.intervals[key].columns.values, cfdna_object2.intervals[key])

    # Merge obs intervals
    for key in cfdna_object2.obs_intervals:
        for obs in cfdna_object2.obs_intervals[key]:
            cfdna_object1.add_obs_intervals(key, obs, cfdna_object2.obs_intervals[key][obs])
    
    # Merge obs values
    for key in cfdna_object2.obs_values:
        cfdna_object1.add_obs_values(key, cfdna_object2.obs_values[key].index.values, cfdna_object2.obs_values[key])

    # Merge HMM states
    for obs in cfdna_object2.hmm_states:
        cfdna_object1.hmm_states[obs] = cfdna_object2.hmm_states[obs]

    # Merge filenames
    cfdna_object1.filenames.update(cfdna_object2.filenames)


def merge_h5(h5_filenames, output_h5, progress_bar=False):
    """
    """

    # Copy first into output
    copy_h5(h5_filenames[0], output_h5)

    # Iterate over h5 files
    for i, h5_filename in enumerate(h5_filenames[1:]):
        # Progress bar
        if progress_bar:
            printProgressBar(i+1, len(h5_filenames), "h5 merge")

        # Read
        cfdna_object1 = cfDNA()
        read_h5(cfdna_object1, output_h5)
        cfdna_object2 = cfDNA()
        read_h5(cfdna_object2, h5_filename)
        # Merge
        merge_cfdna_objects(cfdna_object1, cfdna_object2)
        # Write
        write_h5(cfdna_object1, output_h5)


import ngsfragments as ngs
import numpy as np
from intervalframe import IntervalFrame
import pandas as pd
import os

# Local imports
from ...utilities.h5_utilities import printProgressBar
from .wps import wps


def calculate_nfr(scores, score_name=None, bed_fn="hg19_TSS.bed", upstream=1000, downstream=1000, stranded=False):
    """
    Calculate Nucleosome Free Region (NFR)
    """

    ngs.utilities.window_scores(scores, bed_fn=bed_fn, upstream=upstream, downstream=downstream, stranded=stranded)


def window_mean_scores(frags, values, bed_fn="hg19_TSS.bed", upstream=1000, downstream=1000, stranded=True):
    """
    Calculate mean scores for each postion within a window

    Parameters
    ----------
        values
            numpy.ndarray
        bed_fn
            str
        upstream
            int
        downstream
            int
        stranded
            bool

    Returns
    -------
        mean_values
            numpy.ndarray
    """

    # Check if bed_fn is in data directory
    bed_fn = ngs.data.get_data_file(bed_fn)
    #if os.path.exists(os.path.join(frags.data_dir, bed_fn)):
        #bed_fn = os.path.join(frags.data_dir, bed_fn)
    #else:
        #if ~os.path.exists(bed_fn):
            #raise FileExistsError("bed_fn does not exist!")
    
    # Calculate mean scores
    mean_values = frags.window_mean_scores(values, bed_fn, upstream=upstream, downstream=downstream, stranded=stranded)
    
    return mean_values


def window_nfr(values, bed_fn="hg19_TSS.bed", upstream=1000, downstream=1000, stranded=True, scale=True):
    """
    Calculate Nucleosome Free Region score per window

    Parameters
    ----------
        values
            numpy.ndarray
        bed_fn
            str
        upstream
            int
        downstream
            int
        stranded
            bool
        smooth
            bool
        scale
            bool
        order
            int

    Returns
    -------
        window_values
            pandas.DataFrame
    """

    # Check if bed_fn is in data directory
    bed_fn = ngs.data.get_data_file(bed_fn)
    #if os.path.exists(os.path.join(fragments.data_dir, bed_fn)):
        #bed_fn = os.path.join(fragments.data_dir, bed_fn)
    #else:
        #if ~os.path.exists(bed_fn):
            #raise FileExistsError("bed_fn does not exist!")

    # Calculate scores
    scores = ngs.utilities.window_scores(values, bed_fn=bed_fn, upstream=upstream, downstream=downstream, stranded=stranded)
    
    # Calculate mean scores
    window_values = ngs.utilities.nfr_windows(scores, bed_fn, upstream=upstream, downstream=downstream, scale=scale)

    # Create pandas Series
    window_values = pd.Series(window_values, index=scores.index.values)
    
    return window_values


def predict_nucleosomes(cfdna_object, frags, wps=None, protection=120, merge_distance=5,
                        min_length=120, max_length=210):
    """
    Call peaks from WPS

    Parameters
    ----------
        chrom
            str (chromosomes to call for)

    Returns
    -------
        peaks
            dict[chrom:AIList] (WPS peaks)
    """

    # Calculate wps
    if wps is None:
        peaks = {}
        for chrom in cfdna_object.chroms:
            wps = frags.wps(chrom=chrom, protection=protection, min_length=min_length, max_length=max_length)
            ngs.utilities.normalize_wps(wps)
            tmp_peaks = ngs.utilities.wps_peaks(frags, wps, merge_distance=merge_distance, min_length=min_length, max_length=max_length)
            peaks[chrom] = tmp_peaks[chrom]
    else:
        peaks = ngs.utilities.wps_peaks(frags, wps, merge_distance=merge_distance, min_length=min_length, max_length=max_length)

    return peaks



def peak_distances(cfdna_object, frags, peaks=None, bin_size=100000, max_distance=10000):
    """
    Determine distance between peaks

    Parameters
    ----------
        peaks
            dict[chrom:AIList]
        bin_size
            int
        max_distance
            int
        smooth
            bool

    Returns
    -------
        peak_bins
            dict[pandas.Series]
    """

    # Calculate peaks
    if peaks is None:
        peaks = predict_nucleosomes(cfdna_object, frags)

    # Create p_dist IntervalFrame
    p_dist = IntervalFrame.from_dict_range(frags.genome, bin_size)

    # Iterate over intervals
    p_dist.df.loc[:,"mean_dist"] = 0

    for chrom in peaks:
        o_gen = peaks[chrom].iter_intersect(p_dist.index)
        for i, overlaps in enumerate(o_gen):
            if len(overlaps) > 3:
                p_dist.df.loc[i,"mean_dist"] = np.mean(overlaps.extract_ends() - overlaps.extract_starts())

    # Remove those greater than max_distance
    p_dist = p_dist.iloc[p_dist.df.loc[:,"mean_dist"].values < max_distance, :]

    return p_dist


def summarize_nfr(cfdna_object, frags, key,
                  nfr_filenames=["hg19_TSS.bed.gz","hg19_CTCF.bed.gz","hg19_TFBS.bed.gz"],
                  stranded=[True,False,False], protection=120, min_length=120, max_length=220,
                  cnv_binsize=100000, method="mean", peak_min_length=50, peak_max_length=150,
                  progress_bar=False, verbose=False):
    """
    """

    # Determine file name
    path = os.path.normpath(frags.sam_file)
    file_name = path.split(os.sep)[-1]

    # Check if file was previously annotated
    if file_name not in cfdna_object.filenames:
        cfdna_object.log_filename(frags)

    # Initialize NFR scores
    n = {}
    mean_wps = {}
    for i, bed_fn in enumerate(nfr_filenames):
        title = bed_fn.split(".")[0]
        mean_wps[title] = np.zeros(2000)
        n[title] = 0

    # Determine NFR scores
    peak_dist = IntervalFrame()
    enrichment = {}
    for c, chrom in enumerate(cfdna_object.chroms):
        if progress_bar:
            printProgressBar(c, len(cfdna_object.chroms), "nfr summary " + key)
        if verbose: print(chrom, flush=True)
        wps = frags.wps(chrom=chrom, protection=protection, min_length=min_length, max_length=max_length)
        ngs.utilities.normalize_wps(wps, method=method)

        # Calculate peaks
        if verbose: print("    calling peaks", flush=True)
        peaks = ngs.utilities.wps_peaks(frags, wps, merge_distance=5, min_length=peak_min_length, max_length=peak_max_length)
        if verbose: print("    calling peak distances", flush=True)
        peak_dist = peak_dist.concat([peak_distances(cfdna_object, frags, peaks,
                                                    bin_size=cnv_binsize,
                                                    max_distance=1000)])

        for i, bed_fn in enumerate(nfr_filenames):
            if verbose: print(bed_fn, flush=True)
            title = bed_fn.split(".")[0]
            if verbose: print("    calculating window scores", flush=True)
            scores = ngs.utilities.window_scores(wps, bed_fn=bed_fn, upstream=1000, downstream=1000, stranded=stranded[i], verbose=verbose)
            scores = scores.groupby(by=scores.index.values).mean()
            window_values = ngs.utilities.nfr_windows(scores, scale=True)
            window_values = pd.Series(window_values, index=scores.index.values)

            try:
                enrichment[title]
            except KeyError:
                iterables = [[chrom], window_values.index.values]
                index = pd.MultiIndex.from_product(iterables)
                enrichment[title] = pd.DataFrame(window_values.values, index=index, columns=[file_name])
            try:
                enrichment[title].loc[chrom, :]
                enrichment[title].loc[chrom, file_name] = window_values
            except KeyError:
                iterables = [[chrom], window_values.index.values]
                index = pd.MultiIndex.from_product(iterables)
                new_df = pd.DataFrame(window_values.values, index=index, columns=[file_name])
                enrichment[title] = pd.concat([enrichment[title], new_df])
            
            mean_wps[title] += scores.values.sum(axis=0)
            n[title] += scores.shape[0]

    # Calculate average
    nfr_score = {}
    for title in mean_wps:
        mean_wps[title] /= n[title]
        nfr_score[title] = ngs.utilities.nfr(mean_wps[title], center_x=1000, scale=True, order=35)
    
    # Record all values
    cfdna_object.add_intervals(key + "_peak_dist", file_name, peak_dist)
    for title in mean_wps:
        cfdna_object.add_anno(key + "_" + title + "_nfr_score", file_name, nfr_score[title][0])
        cfdna_object.add_obs_values(key + "_" + title + "_mean_wps", file_name, mean_wps[title])
        cfdna_object.add_obs_values(key + "_" + title + "_nfr_enrichment", file_name, enrichment[title].T)


import numpy as np
from scipy.stats import pearsonr, spearmanr, sem
from scipy.signal import argrelmax
from bokeh.plotting import figure, show, output_file, save
from bokeh.models import Span, ColumnDataSource
from bokeh.transform import linear_cmap
from bokeh.layouts import layout


def fragment_distribution(length_dist, title=None, display=True, save_fn=None):
    """
    Plot the fragment distribution

    Parameters
    ----------
        length_dist : dict
            Dictionary of fragment lengths
        title : str
            Title of plot
        display : bool
            Display plot
        save_fn : str
            Save plot to file

    Returns
    -------
        None
    """
    
    # Determine total distribution
    max_length = max(len(length_dist[chrom]) for chrom in length_dist)
    fragment_lengths = np.zeros(max_length, dtype=int)
    for chrom in length_dist:
        fragment_lengths[:len(length_dist[chrom])] = length_dist[chrom]

    # Creat Axes object if not given
    TOOLS="hover,crosshair,pan,wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,tap,save,box_select,poly_select,lasso_select,"
    p = figure(tools=TOOLS, title=title, x_axis_label="Read length", y_axis_label="Density")
    
    # Plot fragment lengths
    p.line(np.arange(len(fragment_lengths)), fragment_lengths, line_width=2)

    # Plot local maxima
    for maxima in argrelmax(fragment_lengths, order=25)[0]:
        vline = Span(location=maxima, dimension='height', line_color='black', line_width=1)
        p.renderers.extend([vline])
    
    # Save of display plot
    if save_fn is not None:
        output_file(save_fn)
        save(p)
    if display:
        show(p)

    return p


def cnv_plot(cnv_object, title=None, display=True, save_fn=None):
    """
    """

    # Check genome is present
    if cnv_object.genome is None:
        raise AttributeError("Genome file not given")
        
    # Create figure
    TOOLS="hover,crosshair,pan,wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,tap,save,box_select,poly_select,lasso_select,"
    p = figure(tools=TOOLS, title=title, )

    # Create color map
    palette = ['#660220', '#b01b2f', '#d46151', '#f2a585',
               '#fcdbc8', '#f7f7f7', '#d2e5ef', '#94c5dd',
               '#4794c1', '#2668aa', '#083160'][::-1]
    mapper = linear_cmap(field_name='y', palette=palette, low=-0.2, high=0.2)

    # Plot bins
    for chrom in cnv_object.segs:
        chrom_ind = cnv_object.bins.loc[:,"seqnames"].values == chrom
        source = ColumnDataSource(dict(x=cnv_object.bins.loc[chrom_ind,"start"].values + cnv_object.chrom_shift[chrom],
                                        y=cnv_object.bins.loc[chrom_ind,"ratios"].values))

        p.circle(x='x', y='y', line_color=mapper, color=mapper, fill_alpha=1, size=1, source=source)

    # Plot segments
    for chrom in cnv_object.segs:
        # Check that chrom in in genome
        try:
            shift = cnv_object.chrom_shift[chrom]
        except KeyError:
            continue

        # Positive
        for i in cnv_object.segs[chrom]["pos"]:
            p.line([i.start+shift, i.end+shift],
                   [i.value, i.value],
                   color="black")

        # Negative
        for i in cnv_object.segs[chrom]["neg"]:
            p.line([i.start+shift, i.end+shift],
                   [i.value, i.value],
                   color="black")
                    
    # Plot zero lines
    hline = Span(location=0, dimension='width', line_color='darkgrey', line_width=1)
    p.renderers.extend([hline])
    
    # Plot chromosome lines
    last_chrom = list(cnv_object.genome.keys())[-1]
    for chrom in cnv_object.genome:
        vline = Span(location=cnv_object.chrom_shift[chrom], dimension='height', line_color='black', line_width=1)
        p.renderers.extend([vline])
    vline = Span(location=cnv_object.genome[last_chrom] + cnv_object.chrom_shift[last_chrom], dimension='height', line_color='black', line_width=1)
    p.renderers.extend([vline])
                
    # Plot centro lines
    if cnv_object.centro is not None:
        for chrom in cnv_object.genome:
            vline = Span(location=cnv_object.centro[chrom] + cnv_object.chrom_shift[chrom], dimension='height', line_color='grey', line_width=0.75, line_dash="4 4")
            p.renderers.extend([vline])
    
    # Plot chromosome names
    text_x = []
    for chrom in cnv_object.genome:
        text_x.append(int(cnv_object.chrom_shift[chrom] + (cnv_object.genome[chrom] / 2)))
    p.xaxis.ticker = text_x
    p.xaxis.major_label_overrides = {text_x[i]:c for i, c in enumerate(list(cnv_object.genome.keys()))}
    p.xaxis.major_label_orientation = np.pi/2
    p.yaxis.major_label_orientation = "vertical"
    
    # Set attributes
    p.xaxis.bounds = (0, cnv_object.genome[last_chrom] + cnv_object.chrom_shift[last_chrom])
    
    # Save of display plot
    if save_fn != None:
        output_file(save_fn)
        save(p)
    if display:
        show(p)


def plot_profile(frag_profile, genome, chrom_shift, centro, title=None, display=True, save_fn=None):
    """
    """

    # Create figure
    TOOLS="hover,crosshair,pan,wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,tap,save,box_select,poly_select,lasso_select,"
    p = figure(tools=TOOLS, title=title)

    # Plot fragment lengths
    for chrom in frag_profile:
        # Check that chrom in in genome
        try:
            shift = chrom_shift[chrom]
        except KeyError:
            continue

        p.line(frag_profile[chrom].index.values+shift, frag_profile[chrom].values, line_width=2)

    # Plot zero lines
    hline = Span(location=0, dimension='width', line_color='darkgrey', line_width=1)
    p.renderers.extend([hline])
    
    # Plot chromosome lines
    last_chrom = list(genome.keys())[-1]
    for chrom in genome:
        vline = Span(location=chrom_shift[chrom], dimension='height', line_color='black', line_width=1)
        p.renderers.extend([vline])
    vline = Span(location=genome[last_chrom] + chrom_shift[last_chrom], dimension='height', line_color='black', line_width=1)
    p.renderers.extend([vline])
                
    # Plot centro lines
    if centro is not None:
        for chrom in genome:
            vline = Span(location=centro[chrom] + chrom_shift[chrom], dimension='height', line_color='grey', line_width=0.75, line_dash="4 4")
            p.renderers.extend([vline])
    
    # Plot chromosome names
    text_x = []
    for chrom in genome:
        text_x.append(int(chrom_shift[chrom] + (genome[chrom] / 2)))
    p.xaxis.ticker = text_x
    p.xaxis.major_label_overrides = {text_x[i]:c for i, c in enumerate(list(genome.keys()))}
    p.xaxis.major_label_orientation = np.pi/2
    p.yaxis.major_label_orientation = "vertical"
    
    # Set attributes
    p.xaxis.bounds = (0, genome[last_chrom] + chrom_shift[last_chrom])
    
    # Save of display plot
    if save_fn is not None:
        output_file(save_fn)
        save(p)
    if display:
        show(p)

    return p


def plot_dashboard(cnv_object1, cnv_object2, cnv_object3, length_dist, frag_profile, display=True, save_fn=None):
    """
    """

    # Plot length dist
    p1 = fragment_distribution(length_dist, title=None, display=False)

    # Plot cnvs
    p2 = cnv_plot(cnv_object1, title=None, display=False)
    p3 = cnv_plot(cnv_object2, title=None, display=False)
    p4 = cnv_plot(cnv_object3, title=None, display=False)

    # Plot fragment profile
    p5 = plot_profile(frag_profile, cnv_object1.genome, cnv_object1.chrom_shift,
                      cnv_object1.centro, title=None, display=False)

    # Create grid
    l = layout([
                [p1],
                [p2],
                [p3],
                [p4],
                [p5],
               ], sizing_mode="stretch_both")

    # Save of display plot
    if save_fn is not None:
        output_file(save_fn)
        save(l)
    if display:
        show(l)



import numpy as np
import pandas as pd
import ngsfragments as ngs
import os

# Local imports
from ...core.core import cfDNA


def fragment_profile(cfdna_object: cfDNA,
                     frags: ngs.Fragments,
                     bin_size: int = 100000,
                     bin_bias_h5_fn: str = None,
                     bias_correct: bool = True,
                     smooth: bool = True,
                     verbose: bool = False):
   """
   Determine fragment profile (ratio of large fragments to small fragments)

   Parameters
   ----------
      frags : ngsfragments.fragments
      chrom
            str
      bins_size
            int
      genome
            str
      bin_bias_h5_fn
            str
      bias_correct
            bool
      smooth
            bool

   Returns
   -------
      fragment_profile
            pandas.DataFrame
   """

   # Determine file name
   path = os.path.normpath(frags.sam_file)
   file_name = path.split(os.sep)[-1]

   # Check if file was previously annotated
   cfdna_object.log_fragments(frags)

   # Calculate coverage of small and large fragments
   small_bin_coverage = frags.bin_counts(bin_size=bin_size, min_length=100, max_length=150)
   large_bin_coverage = frags.bin_counts(bin_size=bin_size, min_length=151, max_length=220)

   # Correct coverage
   if bias_correct:
      small_bin_coverage = ngs.segment.correction.correct(small_bin_coverage, genome=cfdna_object.ref_genome, bin_size=bin_size, verbose=verbose)
      large_bin_coverage = ngs.segment.correction.correct(large_bin_coverage, genome=cfdna_object.ref_genome, bin_size=bin_size, verbose=verbose)
      small_bin_coverage = small_bin_coverage.loc[:,["corrected_counts"]]
      small_bin_coverage.df.columns = [file_name]
      large_bin_coverage = large_bin_coverage.loc[:,["corrected_counts"]]
      large_bin_coverage.df.columns = [file_name]
   else:
      small_bin_coverage.df.columns = [file_name]
      large_bin_coverage.df.columns = [file_name]

   # Calculate fragment profile
   fragment_profile = small_bin_coverage.copy()
   fragment_profile.df.loc[:,file_name] = small_bin_coverage.df.loc[:,file_name].values / large_bin_coverage.df.loc[:,file_name].values

   # Correct regions with 0 for both
   fragment_profile.df.loc[pd.isnull(fragment_profile.loc[:,file_name].values), file_name] = 1.0
   
   # Append bins
   cfdna_object.add_intervals("frag_profile", fragment_profile)


def fragment_entropy(frags: ngs.Fragments,
                     bin_size: int = 100000,
                     min_length = 100,
                     max_length = 240,
                     n_bins = 5):
      """
      """
      from scipy.stats import entropy

      # Create bins
      cnv = ngs.segment.CNVcaller(cnv_binsize=bin_size)
      bins = cnv.bin_data(frags)
      cbins = cnv.correct_bins(bins)
      shannon_entropy = np.zeros(cbins.shape[0])
      sample = bins.df.columns.values[0]

      # Iterate over chromosomes
      for i, interval in enumerate(cbins.index):
            bin_frags = frags.frags.intersect(interval.start, interval.end, interval.label).filter(min_length, max_length)
            length_dist = np.zeros(max_length-min_length)
            ld = bin_frags.length_dist()[min_length:]
            length_dist[:len(ld)] = ld
            hist_lengths = length_dist.reshape(-1,n_bins).sum(axis=1)
            hist_lengths = hist_lengths / hist_lengths.sum()
            shannon_entropy[i] = entropy(hist_lengths, base=2)

      # Create DataFrame
      cbins.df.columns = ["cnv_ratio"]
      cbins.df.loc[:,"entropy"] = shannon_entropy

      return cbins




import pandas as pd
import numpy as np
import h5py
from projectframe import ProjectFrame

# Local imports
from ...core import cfDNA


def read_h5(h5_filename: str) -> cfDNA:
    """
    Read project frame from HDF5 file

    Parameters
    ----------
        h5_filename : str
            Name of HDF5 file to read from

    Returns
    -------
        cfdna_object : :class: `cfdna.cfDNA`
            cfDNA object
    """

    # Read project frame from HDF5 file
    cfdna_object = ProjectFrame.read_h5(h5_filename)

    return cfdna_object


from .cnv.segmentation import *
from .nucleosome.nfr import *
from .nucleosome.wps import *
from .nucleosome.wps_functions import *
from .fragmentation.frag_pattern import *
from .summarize.summarize import *


__doc__="""\

Tools library for cfDNA analysis
================================
.. currentmodule:: cfdna

Generic
-------
.. autosummary::
   :toctree: .

   nucleosome
   coverage
   fragmentation
   summarize
   cnv

"""


from __future__ import absolute_import
from .core import cfDNA
from . import plot
from . import io
from . import commandline as cmd
from . import tools

from ngsfragments import Fragments


# This is extracted automatically by the top-level setup.py.
__version__ = '2.5.1'

__author__ = "Kyle S. Smith"

__doc__ = """\

API
======

Central class
-------------

.. autosummary::
   :toctree: .
   
   cfDNA


Plotting functions
------------------
    
.. autosummary::
   :toctree: .
   
   plot.plot_plt


IO functions
------------
    
.. autosummary::
   :toctree: .
   
   io.write_h5
   io.write_anno
   io.write_seg
   io.write_anno_seg
   io.read_sam
   io.read_h5


Tools
-----
    
.. autosummary::
   :toctree: .
   
   tools.nucleosome
   tools.cnv

"""


import ngsfragments as ngs
import os
import numpy as np
from typing import List
from ngsfragments import Fragments
from projectframe import ProjectFrame
import ngsfragments as ngs

# Local imports
from ...core.core import cfDNA


def call_cnvs(cfdna_object: cfDNA,
                frags: Fragments,
                cnv_binsize: int = 100000,
                hmm_binsize: int = 1000000,
                genome_version: str = "hg38",
                nthreads: int = 1,
                normal: List[float] = [0.1, 0.25, 0.5, 0.75, 0.9],
                ploidy: List[int] = [2, 3],
                estimatePloidy: bool = False,
                minSegmentBins: int = 25,
                maxCN: int = 5,
                scStates: List[int] = None,
                use_normal: bool = False,
                add_sex: bool = False,
                verbose: bool = False) -> cfDNA:
    """
    Call CNVs from fragments

    This function corrects genome-wide bin coverage for GC, mappability, and black-listed regions.
    Segments are determined using Bayesian Changepoint Segmentation and CNVs and purity  are called 
    using an HMM model. 

    Parameters
    ----------
        cfdna_object : cfDNA
            cfDNA object
        frags : Fragments
            Fragments object
        cnv_binsize : int
            Size of bins for CNV calling (default: 100000)
        hmm_binsize : int
            Size of bins for HMM (default: 1000000)
        genome_version : str
            Reference genome version (default: "hg38")
        nthreads : int
            Number of threads to use (default: 1)
        normal : List[float]
            List of normal fractions (default: [0.1, 0.25, 0.5, 0.75, 0.9])
        ploidy : List[int]
            List of ploidies (default: [2, 3])
        estimatePloidy : bool
            Estimate ploidy (default: False)
        minSegmentBins : int
            Minimum segment bins (default: 25)
        maxCN : int
            Maximum copy number (default: 5)
        scStates : List[int]
            List of states (default: None)
        use_normal : bool
            Use normal (default: False)
        add_sex : bool
            Add sex chromosomes (default: False)
        verbose : bool
            Print process (default: False)
        
    Returns
    -------
        cfdna_object : :class: `cfdna.cfDNA`
            cfDNA object

    Examples
    --------
    >>> import ngsfragments as ngs
    >>> import cfdna
    >>> frags = ngs.io.from_sam("test.bam", genome_version="hg19", nthreads=3)
    >>> cfDNA = cfdna.cfDNA()
    >>> cfDNA = cfdna.tools.cnv.call_cnvs(cfDNA,
                                      frags,
                                      cnv_binsize=100000,
                                      hmm_binsize=1000000,
                                      genome_version="hg19",
                                      nthreads=3,
                                      normal=[0.1, 0.25, 0.5, 0.75, 0.9],
                                      ploidy=[2, 3],
                                      estimatePloidy=False,
                                      minSegmentBins=25,
                                      maxCN=5,
                                      scStates=None,
                                      use_normal=False,
                                      add_sex=True)
    >>> cfDNA

    See Also
    --------
    cfdna.cfDNA : cfDNA object

    """
    
    # Run CNV for hmm models
    prefix = os.path.split(frags.sam_file)[-1].split(".bam")[0]
    cnvs = ngs.segment.CNVcaller(genome_version = genome_version,
                                cnv_binsize = cnv_binsize,
                                hmm_binsize = hmm_binsize,
                                use_normal = use_normal,
                                keep_sex_chroms = add_sex,
                                normal = normal,
                                ploidy = ploidy,
                                estimatePloidy = estimatePloidy,
                                scStates = scStates,
                                minSegmentBins = minSegmentBins,
                                maxCN = maxCN)
        
    # Call CNVs
    sample = os.path.split(prefix)[-1]
    cnvs.predict_cnvs(frags)
    cfdna_object = cfDNA(pf=cnvs.pf, genome_version=genome_version)
    cfdna_object.log_fragments(frags)
    cfdna_object.params["cnv_binsize"] = cnv_binsize
    # Calculate segment variance
    cfdna_object.obs_intervals[sample]["cnv_segments"].annotate(cfdna_object.intervals["cnv_bins"], sample, "var")
    
    return cfdna_object


from .read.read_bam import *
from .read.read_h5 import *
from .write.write_h5 import *
from .write.write_text import *

__doc__="""\

IO API
============
.. currentmodule:: cfdna.io

Functions
---------
.. autosummary::
   :toctree: .

   read_bam
   read_h5
   write_h5
   write_seg

"""


"""
Create matplotlib plots from cfDNA object
"""

import numpy as np
import matplotlib.pyplot as plt
import ngsfragments as ngs
import pandas as pd
from ..core.core import cfDNA

# Tell matplot lib to use 'editable text'
import matplotlib
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42


def cnv_summary(cfdna_object: cfDNA,
                sample: str,
                add_wps: bool = False,
                show: bool = True,
                save: str = None) -> plt.figure:
    """
    Plot CNV summary

    Parameters
    ----------
        cfdna_object : cfDNA
            cfDNA object
        sample : str
            Sample name
        add_wps : bool
            Add WPS (default: False)
        show : bool
            Show plot (default: True)
        save : str
            Save plot to file (default: None)
    
    Returns
    -------
        fig : plt.figure
            matplotlib.pyplot.figure
    """

    # Create figure
    fig = ngs.plot.cnv_summary(cfdna_object,
                            sample,
                            add_wps = add_wps,
                            show = show,
                            save = save)

    return fig


from .segmentation import *


__doc__="""\

Copy Number Variation analysis
==============================
.. currentmodule:: cfdna.tools.cnv

Generic
-------
.. autosummary::
   :toctree: .

   call_cnvs

"""


from ngsfragments import Fragments
import ngsfragments as ngs


def read_sam(bam_filename: str,
            min_size: int = 10,
            max_size: int = 1000,
            paired: bool = True,
            qcfail: bool = False,
            mapq_cutoff: int = 25,
            proportion: float = 1.0,
            n_frags: int = None,
            nthreads: int = 1,
            verbose: bool = False) -> Fragments:
    """
    Read BAM file

    Parameters
    ----------
        bam_filename : str
            Name of the BAM file to read
        min_size : int
            Minimum length of fragment to consider
        max_size : int
            Maximum length of fragment to consider
        paired : bool
            Whether to process as paired reads
        qcfail : bool
            Whether to keep QC failed reads
        mapq_cutoff : int
            Minimum mapping quality allowed
        proportion : float
            Proportion of reads to keep
        n_frags : int
            Number of fragments to keep
        nthreads : int
            Number of threads to use
        verbose : bool
            Print process
    
    Returns
    -------
        fragments : :class: `ngsfragments.ngsfragments`
            DNA fragment intervals from BAM file

    """

    # Read fragments from bam file
    frags = ngs.read_sam.from_sam(bam_filename,
                                    min_size = min_size,
                                    max_size = max_size,
                                    paired = paired,
                                    qcfail = qcfail,
                                    mapq_cutoff = mapq_cutoff,
                                    verbose = verbose,
                                    proportion = proportion,
                                    n_frags = n_frags,
                                    nthreads = nthreads)

    return frags


from ..core.core import cfDNA
import os
import multiprocessing
import matplotlib.pyplot as plt
import numpy as np
import ngsfragments as ngs
from intervalframe import IntervalFrame
from ailist import LabeledIntervalArray
import genome_info
import glob

# Local imports
from ..data.import_data import get_data_file
from ..tools.nucleosome.wps import wps_gene_fft
from ..tools.nucleosome.wps import predict_nucleosomes


def CNV_calling(args):
    """
    """

    # Check number of cores
    if args.nthreads == -1:
        nthreads = multiprocessing.cpu_count()
    else:
        nthreads = args.nthreads

    # Detect if multiple bams
    if isinstance(args.bam, list):
        bams = args.bam
    else:
        bams = [args.bam]

    # Read PON
    if args.pon != "":
        pon = IntervalFrame.read_parquet(args.pon)

    # Read additional blacklist
    if args.genome == "hg38":
        blacklist = IntervalFrame.read_parquet(get_data_file("cfdna_hg38_blacklist.parquet"))
    else:
        blacklist = None

    for bam in bams:
        if args.prefix == "":
            prefix = os.path.split(bam)[-1].split(".bam")[0]
        else:
            prefix = args.prefix

        # Create cfDNA object
        if args.cache != "":
            # Check if cache exists
            if os.path.exists(args.cache):
                cfdna_object = cfDNA.from_h5(args.cache)
        else:
            cfdna_object = cfDNA(genome_version=args.genome)

            # Read bam
            if args.nanopore:
                import pysam
                samfile = pysam.AlignmentFile(bam, "rb")
                ail = LabeledIntervalArray()
                iters = samfile.fetch(until_eof=True)
                for x in iters:
                    if x.is_unmapped == False and x.is_secondary == False and x.is_supplementary == False:
                        length = x.reference_end - x.reference_start
                        if length >= args.min_length and length <= args.max_length:
                            ail.add(int(x.reference_start), int(x.reference_end), x.reference_name)
                frags = ngs.Fragments(genome_version=args.genome)
                frags.frags = ail
                frags.sam_file = bam
            else:
                frags = ngs.io.from_sam(bam,
                                        genome_version=args.genome,
                                        min_size = args.min_length,
                                        max_size = args.max_length,
                                        paired = not args.single,
                                        qcfail = args.qcfail,
                                        mapq_cutoff = args.mapq,
                                        nthreads=nthreads,
                                        verbose=args.verbose)
            
            # Run CNV for hmm models
            if args.clonal:
                cnvs = ngs.segment.CNVcaller(genome_version=args.genome,
                                            cnv_binsize=args.bin_size,
                                            hmm_binsize=args.hmm_bin_size,
                                            use_normal = args.use_normal,
                                            keep_sex_chroms = args.add_sex,
                                            normal = [0.1, 0.25, 0.5, 0.75, 0.9],
                                            ploidy = [2, 3, 4],
                                            estimatePloidy = True,
                                            scStates = [1, 3],
                                            minSegmentBins = 25,
                                            maxCN = 5,
                                            rpca_normal = args.rpca_normal,
                                            n_mads = args.n_mads,
                                            bcp_cutoff = args.bcp_cutoff)
                
            else:
                cnvs = ngs.segment.CNVcaller(genome_version=args.genome,
                                            cnv_binsize=args.bin_size,
                                            hmm_binsize=args.hmm_bin_size,
                                            use_normal = args.use_normal,
                                            keep_sex_chroms = args.add_sex,
                                            normal = [0.1, 0.25, 0.5, 0.75, 0.9],
                                            ploidy = [2, 3, 4],
                                            estimatePloidy = True,
                                            scStates = None,
                                            minSegmentBins = 25,
                                            maxCN = 5,
                                            rpca_normal = args.rpca_normal,
                                            n_mads = args.n_mads,
                                            bcp_cutoff = args.bcp_cutoff)
            # Call CNVs
            sample = os.path.split(prefix)[-1]
            if args.pon != "":
                cnvs.predict_cnvs(frags, normal_data=pon, additional_blacklist=blacklist, additional_blacklist_cutoff=0.1, merge_MAD=args.n_mads)
            else:
                cnvs.predict_cnvs(frags, additional_blacklist=blacklist, additional_blacklist_cutoff=0.1, merge_MAD=args.n_mads)
            cfdna_object = cfDNA(pf=cnvs.pf, genome_version=args.genome)
            cfdna_object.log_fragments(frags)
            cfdna_object.params["cnv_binsize"] = args.bin_size
            # Calculate segment variance
            cfdna_object.obs_intervals[sample]["cnv_segments"].annotate(cfdna_object.intervals["cnv_bins"], sample, "var")
            
            #cfdna_object.add_anno("n_fragments", sample, len(frags.frags))
            # Calculate segment variance
            #cfdna_object.obs_intervals[sample]["cnv_segments"].annotate(cfdna_object.intervals["cnv_bins"], sample, "var")
            #length_dist = frags.length_dist()
            #length_dist.name = sample
            #cfdna_object.add_values("length_dist", length_dist)
            
            # Add WPS
            if args.add_wps:
                scores = ngs.metrics.wps_windows(frags,
                                                protection = 120,
                                                min_length = args.min_length,
                                                max_length = args.max_length,
                                                feature = "tss",
                                                smooth = False)
                cfdna_object.add_obs_values(list(cfdna_object.obs)[0], "tss_wps", scores)
        
        # Write seg file
        if args.segs:
            seg_fn = prefix + ".seg"
            sample = os.path.split(prefix)[-1]
            ngs.segment.cnv_utilities.write_seg_file(cfdna_object,
                                                        seg_fn,
                                                        sample)

        # Plot
        ngs.plot.cnv_summary(cfdna_object,
                            list(cfdna_object.obs)[0],
                            add_wps = args.add_wps,
                            show = False,
                            save = prefix+"_cnv_plot.pdf")
        
        # Write h5
        if args.cache != "" and os.path.exists(args.cache) == False:
            cfdna_object.to_h5(args.cache)

        # Write annotations
        if args.anno_file:
            cfdna_object.anno.engine.df.to_csv(prefix+"_metrics.txt", header=True, index=True, sep="\t")
        
        # Write seg annotations
        if args.anno_segs:
            sample = os.path.split(prefix)[-1]
            df = cfdna_object.obs_intervals[sample]["cnv_segments"].df
            df.loc[:,"chrom"] = cfdna_object.obs_intervals[sample]["cnv_segments"].index.labels
            df.loc[:,"start"] = cfdna_object.obs_intervals[sample]["cnv_segments"].index.starts
            df.loc[:,"end"] = cfdna_object.obs_intervals[sample]["cnv_segments"].index.ends
            df.loc[:,"sample"] = sample
            df.loc[:,"n_bins"] = ((df.loc[:,"end"].values - df.loc[:,"start"].values) / args.bin_size).astype(int)
            df = df.loc[:,['sample', 'chrom', 'start', 'end', 'copy_number', 'event', 'subclone_status',
                           'logR_Copy_Number', 'Corrected_Copy_Number', 'Corrected_Call', 'var', 'n_bins', 'median']]
            df.to_csv(prefix+"_seg_annotations.seg", header=True, index=False, sep="\t")

            # Drop columns
            df.drop(columns=["chrom", "start", "end"], inplace=True)

        # Write bins
        if args.bins_file:
            sample = os.path.split(prefix)[-1]
            df = cfdna_object.intervals["cnv_bins"].df
            df.loc[:,"chrom"] = cfdna_object.intervals["cnv_bins"].index.labels
            df.loc[:,"start"] = cfdna_object.intervals["cnv_bins"].index.starts
            df.loc[:,"end"] = cfdna_object.intervals["cnv_bins"].index.ends
            df.loc[:,"ratio"] = df.loc[:,sample].values
            df = df.loc[:,['chrom', 'start', 'end', 'ratio']]
            df.to_csv(prefix+"_bins.txt", header=True, index=False, sep="\t")

            # Drop columns
            df.drop(columns=["chrom", "start", "end", "ratio"], inplace=True)


def GeneActivity(args):
    """
    """

    # Check number of cores
    if args.nthreads == -1:
        nthreads = multiprocessing.cpu_count()
    else:
        nthreads = args.nthreads

    # Detect if multiple bams
    if isinstance(args.bam, list):
        bams = args.bam
    else:
        bams = [args.bam]

    for bam in bams:
        if args.prefix == "":
            prefix = os.path.split(bam)[-1].split(".bam")[0]
        else:
            prefix = args.prefix
        sample = os.path.split(prefix)[-1]
        
        # Read bam
        frags = ngs.io.from_sam(bam,
                                genome_version=args.genome,
                                min_size = args.min_length,
                                max_size = args.max_length,
                                paired = not args.single,
                                qcfail = args.qcfail,
                                mapq_cutoff = args.mapq,
                                nthreads=nthreads,
                                verbose=args.verbose)

        # Create cfDNA object
        if args.cache != "":
            cfdna_object = cfDNA.from_h5(args.cache)
        else:
            cfdna_object = cfDNA(genome_version=args.genome)

        # Get gene activity
        gene_pattern = wps_gene_fft(frags,
                                    genome_version = args.genome,
                                    protection = args.protection,
                                    min_length = args.min_length,
                                    max_length = args.max_length,
                                    freq_range = (120, 280),
                                    scale = True,
                                    feature = args.feature,
                                    verbose = args.verbose)
        #gene_pattern = gene_pattern.loc[gene_pattern.sum(axis=1).values != 0,:]
        gene_activity = gene_pattern.loc[:,193:205].mean(axis=1).to_frame()
        gene_activity.columns = [sample]

        # Add to cfDNA object
        cfdna_object.add_values("gene_activity", gene_activity)

    # Write parquet
    cfdna_object.values["gene_activity"].to_parquet("gene_activity.parquet")


def GenomeCoverage(args):
    """
    """
    import pyBigWig

    # Check number of cores
    if args.nthreads == -1:
        nthreads = multiprocessing.cpu_count()
    else:
        nthreads = args.nthreads

    # Detect if multiple bams
    if isinstance(args.bam, list):
        bams = args.bam
    else:
        bams = [args.bam]

    for bam in bams:
        if args.prefix == "":
            prefix = os.path.split(bam)[-1].split(".bam")[0]
        else:
            prefix = args.prefix
        sample = os.path.split(prefix)[-1]
        
        # Read bam
        frags = ngs.io.from_sam(bam,
                                genome_version=args.genome,
                                min_size = args.min_length,
                                max_size = args.max_length,
                                paired = not args.single,
                                qcfail = args.qcfail,
                                mapq_cutoff = args.mapq,
                                nthreads=nthreads,
                                verbose=args.verbose)
        # Calculate coverage
        bw = pyBigWig.open(sample+"_coverage.bw", "w")
        header = []
        chrom_ranges = frags.frags.label_ranges
        for chrom in chrom_ranges:
            ranges = chrom_ranges[chrom]
            header.append((chrom, ranges[1]))
        bw.addHeader(header)
        chroms = frags.frags.unique_labels
        for chrom in chroms:
            if args.verbose: print(chrom)
            cov = ngs.coverage.coverage(frags, chrom=chrom, min_length=args.min_length, max_length=args.max_length)
            ngs.coverage.normalize_coverage(cov, method="mean")
            cov[chrom] = cov[chrom].astype(np.float32)

            # Write to bigwig
            bw.addEntries(chrom, cov[chrom].index.values, values=cov[chrom].values, span=1)
        
        bw.close()


def GenomeWPS(args):
    """
    """
    import pyBigWig

    # Check number of cores
    if args.nthreads == -1:
        nthreads = multiprocessing.cpu_count()
    else:
        nthreads = args.nthreads

    # Detect if multiple bams
    if isinstance(args.bam, list):
        bams = args.bam
    else:
        bams = [args.bam]

    for bam in bams:
        if args.prefix == "":
            prefix = os.path.split(bam)[-1].split(".bam")[0]
        else:
            prefix = args.prefix
        sample = os.path.split(prefix)[-1]
        
        # Read bam
        frags = ngs.io.from_sam(bam,
                                genome_version=args.genome,
                                min_size = args.min_length,
                                max_size = args.max_length,
                                paired = not args.single,
                                qcfail = args.qcfail,
                                mapq_cutoff = args.mapq,
                                nthreads=nthreads,
                                verbose=args.verbose)
        # Calculate coverage
        bw = pyBigWig.open(sample+"_wps.bw", "w")
        header = []
        chrom_ranges = frags.frags.label_ranges
        for chrom in chrom_ranges:
            ranges = chrom_ranges[chrom]
            header.append((chrom, ranges[1]))
        bw.addHeader(header)
        chroms = frags.frags.unique_labels
        for chrom in chroms:
            if args.verbose: print(chrom)
            cov = ngs.coverage.wps(frags, chrom=chrom, min_length=args.min_length, max_length=args.max_length, protection=args.protection)
            ngs.coverage.normalize_coverage(cov, method="mean")
            cov[chrom] = cov[chrom].astype(np.float32)

            # Write to bigwig
            bw.addEntries(chrom, cov[chrom].index.values, values=cov[chrom].values, span=1)
        
        bw.close()


def WPSpeaks(args):
    """
    """
    import pyBigWig

    # Check number of cores
    if args.nthreads == -1:
        nthreads = multiprocessing.cpu_count()
    else:
        nthreads = args.nthreads

    # Detect if multiple bams
    if isinstance(args.bam, list):
        bams = args.bam
    else:
        bams = [args.bam]

    for bam in bams:
        if args.prefix == "":
            prefix = os.path.split(bam)[-1].split(".bam")[0]
        else:
            prefix = args.prefix
        sample = os.path.split(prefix)[-1]
        
        # Read bam
        frags = ngs.io.from_sam(bam,
                                genome_version=args.genome,
                                min_size = args.min_length,
                                max_size = args.max_length,
                                paired = not args.single,
                                qcfail = args.qcfail,
                                mapq_cutoff = args.mapq,
                                nthreads=nthreads,
                                verbose=args.verbose)
        
        peaks = predict_nucleosomes(frags,
                        protection = args.protection,
                        merge_distance = 5,
				        min_length = args.min_length,
                        max_length = args.max_length)
        iframe_peaks = IntervalFrame(intervals=peaks)
        iframe_peaks.to_bed(prefix+"_nucleosomes.bed")


def create_pon(args):
    """
    """

    # Load genome
    g = genome_info.GenomeInfo(args.genome)

    # Check number of cores
    if args.nthreads == -1:
        nthreads = multiprocessing.cpu_count()
    else:
        nthreads = args.nthreads

    # Detect if multiple bams
    bams = glob.glob(args.dir + "/*.bam")

    # Create pon
    pon = LabeledIntervalArray.create_bin(g["chrom_sizes"], bin_size=args.bin_size)
    pon = IntervalFrame(intervals=pon)
    pon = pon.loc[g["main_chromosomes"],:]

    for bam in bams:
        if args.verbose: print(bam)
        prefix = os.path.split(bam)[-1].split(".bam")[0]
        sample = os.path.split(prefix)[-1]
        
        # Read bam
        frags = ngs.io.from_sam(bam,
                                genome_version=args.genome,
                                min_size = args.min_length,
                                max_size = args.max_length,
                                paired = not args.single,
                                qcfail = args.qcfail,
                                mapq_cutoff = args.mapq,
                                nthreads=nthreads,
                                verbose=args.verbose)

        # Bin coverage
        cnv = ngs.segment.CNVcaller(genome_version=args.genome, cnv_binsize=args.bin_size)
        bins = cnv.bin_data(frags)
        bins = bins.loc[g["main_chromosomes"],:]
        bins = bins.exact_match(pon)
        pon = pon.exact_match(bins)

        # Calculate coverage
        pon.df.loc[:,sample] = bins.df.loc[:,sample].values

    # Write to parquet
    pon.to_parquet(args.out)


def TFwps(args):
    """
    """

    # Check number of cores
    if args.nthreads == -1:
        nthreads = multiprocessing.cpu_count()
    else:
        nthreads = args.nthreads

    # Detect if multiple bams
    if isinstance(args.bam, list):
        bams = args.bam
    else:
        bams = [args.bam]

    for bam in bams:
        if args.prefix == "":
            prefix = os.path.split(bam)[-1].split(".bam")[0]
        else:
            prefix = args.prefix
        sample = os.path.split(prefix)[-1]

        # Read bam
        frags = ngs.io.from_sam(bam,
                                genome_version=args.genome,
                                min_size = args.min_length,
                                max_size = args.max_length,
                                paired = not args.single,
                                qcfail = args.qcfail,
                                mapq_cutoff = args.mapq,
                                nthreads=nthreads,
                                verbose=args.verbose)
        tf_values = ngs.metrics.wps_metrics.wps_score_tfs(frags, protection=args.protection, min_length=args.min_length, max_length=args.max_length, upstream=5000, downstream=5000)
        tf_values.columns = np.arange(-5000,5001)
        tf_values.to_parquet(prefix+"_tf_wps.parquet")


import argparse
from .commandline.commands import CNV_calling, GeneActivity, GenomeWPS, GenomeCoverage, create_pon


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    # Call CNVs
    parser_cnv = subparsers.add_parser("callCNVs", help="Plot and write CNV segments")
    parser_cnv.add_argument("--bam", help="BAM file", required=True, nargs='*')
    parser_cnv.add_argument("--pon", help="Panel of normals parquet generated from createPON", default="")
    parser_cnv.add_argument("--prefix", help="Prefix for ouput files", default="")
    parser_cnv.add_argument("--bin_size", type=int, help="Bin size to use (default=100000)", default=100000)
    parser_cnv.add_argument("--hmm_bin_size", type=int, help="Bin size to use (default=1000000)", default=1000000)
    parser_cnv.add_argument("--genome", help="Version of genome to use (default=hg38)", default="hg38")
    parser_cnv.add_argument("--proportion", type=float, help="Proportion of fragments to use (default: 1.0)", default=1.0)
    parser_cnv.add_argument("--n_frags", type=int, help="Estimate of number of fragments to use (default: ALL)", default=0)
    parser_cnv.add_argument("--min_length", type=int, help="Minimum length to consider (default: 1)", default=1)
    parser_cnv.add_argument("--max_length", type=int, help="Maximum length to consider (default: 1000)", default=1000)
    parser_cnv.add_argument("--mapq", type=int, help="Mapping quality cutoff (default: 25)", default=25)
    parser_cnv.add_argument("--segs", help="Whether to write seg files (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--nthreads", type=int, help="Number of threads to use (default=1)", default=1)
    parser_cnv.add_argument("--n_mads", type=float, help="Number of MADs to merge segments (default: 1.4826)", default=1.4826)
    parser_cnv.add_argument("--bcp_cutoff", type=float, help="Cutoff for bayesian changpoint segmentation (default: 0.3)", default=0.3)
    parser_cnv.add_argument("--cache", help="Name of h5 file to append results to", default="")
    parser_cnv.add_argument("--single", help="Whether to reads are single ended (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--qcfail", help="Whether to remove qcfail flag (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--use_normal", help="Whether to correct using normal profiles (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--add_wps", help="Whether to add a WPS plot for TSSs (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--add_sex", help="Whether to keep sex chromosomes (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--clonal", help="Whether to predict clonality (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--anno_file", help="Whether to write text file with predictioned metrics (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--anno_segs", help="Whether to annotated seg file (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--bins_file", help="Whether to write bins file (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--rpca_normal", help="Whether to use rPCA for normal correction (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--nanopore", help="Whether input bam is from nanopore (default: False)", default=False, action="store_true")
    parser_cnv.add_argument("--verbose", help="Whether to be verbose (default: False)", default=False, action="store_true")
    parser_cnv.set_defaults(func=CNV_calling)

    # Call Gene activity
    parser_gene = subparsers.add_parser("geneActivity", help="Write gene activity")
    parser_gene.add_argument("--bam", help="BAM file", required=True, nargs='*')
    parser_gene.add_argument("--prefix", help="Prefix for ouput files", default="")
    parser_gene.add_argument("--genome", help="Version of genome to use (default=hg38)", default="hg38")
    parser_gene.add_argument("--feature", help="Feature to use (default=gene)", default="gene")
    parser_gene.add_argument("--protection", type=int, help="Protection winodw size (default: 120)", default=120)
    parser_gene.add_argument("--min_length", type=int, help="Minimum length to consider (default: 120)", default=120)
    parser_gene.add_argument("--max_length", type=int, help="Maximum length to consider (default: 220)", default=220)
    parser_gene.add_argument("--mapq", type=int, help="Mapping quality cutoff (default: 25)", default=25)
    parser_gene.add_argument("--nthreads", type=int, help="Number of threads to use (default=1)", default=1)
    parser_gene.add_argument("--cache", help="Name of h5 file to append results to", default="")
    parser_gene.add_argument("--single", help="Whether to reads are single ended (default: False)", default=False, action="store_true")
    parser_gene.add_argument("--qcfail", help="Whether to remove qcfail flag (default: False)", default=False, action="store_true")
    parser_gene.add_argument("--verbose", help="Whether to be verbose (default: False)", default=False, action="store_true")
    parser_gene.set_defaults(func=GeneActivity)

    # Call Genome WPS
    parser_wps = subparsers.add_parser("genomeWPS", help="Write genome WPS")
    parser_wps.add_argument("--bam", help="BAM file", required=True, nargs='*')
    parser_wps.add_argument("--prefix", help="Prefix for ouput files", default="")
    parser_wps.add_argument("--genome", help="Version of genome to use (default=hg38)", default="hg38")
    parser_wps.add_argument("--protection", type=int, help="Protection winodw size (default: 120)", default=120)
    parser_wps.add_argument("--min_length", type=int, help="Minimum length to consider (default: 120)", default=120)
    parser_wps.add_argument("--max_length", type=int, help="Maximum length to consider (default: 220)", default=220)
    parser_wps.add_argument("--mapq", type=int, help="Mapping quality cutoff (default: 25)", default=25)
    parser_wps.add_argument("--nthreads", type=int, help="Number of threads to use (default=1)", default=1)
    parser_wps.add_argument("--single", help="Whether to reads are single ended (default: False)", default=False, action="store_true")
    parser_wps.add_argument("--qcfail", help="Whether to remove qcfail flag (default: False)", default=False, action="store_true")
    parser_wps.add_argument("--verbose", help="Whether to be verbose (default: False)", default=False, action="store_true")
    parser_wps.set_defaults(func=GenomeWPS)

    # Call Genome coverage
    parser_cov = subparsers.add_parser("genomeCov", help="Write genome coverage")
    parser_cov.add_argument("--bam", help="BAM file", required=True, nargs='*')
    parser_cov.add_argument("--prefix", help="Prefix for ouput files", default="")
    parser_cov.add_argument("--genome", help="Version of genome to use (default=hg38)", default="hg38")
    parser_cov.add_argument("--min_length", type=int, help="Minimum length to consider (default: 1)", default=1)
    parser_cov.add_argument("--max_length", type=int, help="Maximum length to consider (default: 1000)", default=1000)
    parser_cov.add_argument("--mapq", type=int, help="Mapping quality cutoff (default: 25)", default=25)
    parser_cov.add_argument("--nthreads", type=int, help="Number of threads to use (default=1)", default=1)
    parser_cov.add_argument("--single", help="Whether to reads are single ended (default: False)", default=False, action="store_true")
    parser_cov.add_argument("--qcfail", help="Whether to remove qcfail flag (default: False)", default=False, action="store_true")
    parser_cov.add_argument("--verbose", help="Whether to be verbose (default: False)", default=False, action="store_true")
    parser_cov.set_defaults(func=GenomeCoverage)

    # Call create PON
    parser_pon = subparsers.add_parser("createPON", help="Create panel of normals")
    parser_pon.add_argument("--dir", help="Directory with BAM files", required=True)
    parser_pon.add_argument("--out", help="Ouput file name", required=True)
    parser_pon.add_argument("--genome", help="Version of genome to use (default=hg38)", default="hg38")
    parser_pon.add_argument("--bin_size", type=int, help="Bin size (default: 10000)", default=10000)
    parser_pon.add_argument("--min_length", type=int, help="Minimum length to consider (default: 1)", default=1)
    parser_pon.add_argument("--max_length", type=int, help="Maximum length to consider (default: 1000)", default=1000)
    parser_pon.add_argument("--mapq", type=int, help="Mapping quality cutoff (default: 25)", default=25)
    parser_pon.add_argument("--nthreads", type=int, help="Number of threads to use (default=1)", default=1)
    parser_pon.add_argument("--single", help="Whether to reads are single ended (default: False)", default=False, action="store_true")
    parser_pon.add_argument("--qcfail", help="Whether to remove qcfail flag (default: False)", default=False, action="store_true")
    parser_pon.add_argument("--verbose", help="Whether to be verbose (default: False)", default=False, action="store_true")
    parser_pon.set_defaults(func=create_pon)

    # Call TF WPS
    parser_wps = subparsers.add_parser("TFwps", help="Write TF WPS")
    parser_wps.add_argument("--bam", help="BAM file", required=True, nargs='*')
    parser_wps.add_argument("--prefix", help="Prefix for ouput files", default="")
    parser_wps.add_argument("--genome", help="Version of genome to use (default=hg38)", default="hg38")
    parser_wps.add_argument("--protection", type=int, help="Protection winodw size (default: 60)", default=60)
    parser_wps.add_argument("--min_length", type=int, help="Minimum length to consider (default: 60)", default=60)
    parser_wps.add_argument("--max_length", type=int, help="Maximum length to consider (default: 1000)", default=1000)
    parser_wps.add_argument("--mapq", type=int, help="Mapping quality cutoff (default: 25)", default=25)
    parser_wps.add_argument("--nthreads", type=int, help="Number of threads to use (default=1)", default=1)
    parser_wps.add_argument("--single", help="Whether to reads are single ended (default: False)", default=False, action="store_true")
    parser_wps.add_argument("--qcfail", help="Whether to remove qcfail flag (default: False)", default=False, action="store_true")
    parser_wps.add_argument("--verbose", help="Whether to be verbose (default: False)", default=False, action="store_true")
    parser_wps.set_defaults(func=GenomeWPS)

    args = parser.parse_args()

    args.func(args)


import os


def get_data_file(filename):
    """
    """

    # Check if filename already exists
    #if os.path.exists(filename):

    # Find data directory
    data_dir = os.path.split(os.path.realpath(__file__))[0]
    data_dir = os.path.join(data_dir, filename)

    # Check if file exists
    if os.path.exists(data_dir) == False:
        raise FileExistsError("Data file does not exist! ("+data_dir+")")

    return data_dir


from .nfr import *
from .wps import *


__doc__="""\

Nucleosome positioning analysis
===============================
.. currentmodule:: cfdna.tools.nucleosome

Generic
-------
.. autosummary::
   :toctree: .

   wps_windows

"""


import numpy as np
import pandas as pd
import ngsfragments as ngs
import os


# Local imports
from ...core.core import cfDNA


def write_seg(cfdna_object: cfDNA,
              prefix: str) -> None:
    """
    Write seg file

    Parameters
    ----------
        cfdna_object : cfDNA
            cfDNA object
        prefix : str
            Prefix for output file

    Returns
    -------
        None

    """
    
    # Write seg file
    seg_fn = prefix + ".seg"
    sample = os.path.split(prefix)[-1]
    ngs.segment.cnv_utilities.write_seg_file(cfdna_object,
                                             seg_fn,
                                             sample)
    
    return None


def write_anno(cfdna_object: cfDNA,
               prefix: str) -> None:
    """
    Write annotation file

    Parameters
    ----------
        cfdna_object : cfDNA
            cfDNA object
        prefix : str
            Prefix for output file

    Returns
    -------
        None

    
    """

    # Write annotation file
    cfdna_object.anno.engine.df.to_csv(prefix+"_metrics.txt", header=True, index=True, sep="\t")

    return None


def write_anno_seg(cfdna_object: cfDNA,
                   prefix: str,
                   bin_size: int) -> None:
    """
    Write annotationed seg file

    Parameters
    ----------
        cfdna_object : cfDNA
            cfDNA object
        prefix : str
            Prefix for output file
        bin_size : int
            Bin size
    
    Returns
    -------
        None
    """

    # Write annotation file
    sample = os.path.split(prefix)[-1]
    df = cfdna_object.obs_intervals[sample]["cnv_segments"].df
    df.loc[:,"chrom"] = cfdna_object.obs_intervals[sample]["cnv_segments"].index.labels
    df.loc[:,"start"] = cfdna_object.obs_intervals[sample]["cnv_segments"].index.starts
    df.loc[:,"end"] = cfdna_object.obs_intervals[sample]["cnv_segments"].index.ends
    df.loc[:,"sample"] = sample
    df.loc[:,"n_bins"] = ((df.loc[:,"end"].values - df.loc[:,"start"].values) / bin_size).astype(int)
    df = df.loc[:,['sample', 'chrom', 'start', 'end', 'copy_number', 'event', 'subclone_status',
                    'logR_Copy_Number', 'Corrected_Copy_Number', 'Corrected_Call', 'var', 'n_bins', 'median']]
    df.to_csv(prefix+"_seg_annotations.seg", header=True, index=False, sep="\t")

    # Drop columns
    df.drop(columns=["chrom", "start", "end"], inplace=True)

    return None


import os
from ngsfragments import Fragments
from dataclasses import field
from projectframe import ProjectFrame, Frame, MultiFrame, MultiIntervalFrame, UnstructuredLookup, ObsFrame, ObsIntervalFrame


# Local imports

class cfDNA(ProjectFrame):
    """
    Wrapper of ProjectFrame for cell-free DNA

	:class:`~cfdna.cfDNA` stores a fragments object

    """

    def __init__(self,
                 anno: Frame = None,
                 values: MultiFrame = None,
                 obs_values: ObsFrame = None,
                 intervals: MultiFrame  = None,
                 obs_intervals: ObsIntervalFrame  = None,
                 obs: set = None,
                 uns: UnstructuredLookup = None,
                 params: UnstructuredLookup = None,
                 genome_version: str = "hg19",
                 pf: ProjectFrame = None):
        """
        Initialize cfDNA object
                 
        Parameters
        ----------
            anno : Frame
                Annotation table
            values : MultiFrame
                Data table
            obs_values : ObsFrame
                Observed data table
            intervals : MultiFrame
                Interval table
            obs_intervals : ObsIntervalFrame
                Observed interval table
            obs : set
                Set of observed samples
            uns : UnstructuredLookup
                Unstructured data
            params : UnstructuredLookup
                Parameters
            genome_version : str
                Reference genome version (default: "hg19")
            pf : ProjectFrame
                Inpute projectframe
                 
        Returns
        -------
            None

        """

        if pf is None:
            super().__init__(anno, values, obs_values, intervals,
                            obs_intervals, obs, uns, params)
        else:
            super().__init__(pf.anno, pf.values, pf.obs_values, pf.intervals,
                            pf.obs_intervals, pf.obs, pf.uns, pf.params)
        
        self.params["ref_genome"] = genome_version


    def log_fragments(self,
                      frags: Fragments) -> None:
        """
        Log Fragments object to ProjectFrame

        Parameters
        ----------
            frags : Fragments
                Fragments object

        Returns
        -------
            None

        Examples
        --------
        >>> import ngsfragments as ngs
        >>> import cfdna
        >>> frags = ngs.io.read_sam("test.bam", nthreads=3, genome_version="hg19")
        >>> cfDNA = cfdna.cfDNA()
        >>> cfDNA.log_fragments(frags)
        """

        # Determine name
        file_name = frags.sam_file
        sample_name = os.path.split(file_name)[-1].split(".bam")[0]

        # Check is file_name exists
        #if sample_name not in pf.obs:
        self.add_obs(sample_name)

        # Record
        if "length_dist" not in self.values.keys or pd.isnull(self.values["length_dist"].loc[sample_name,:].values).all():
            length_dist = frags.length_dist()
            length_dist.name = sample_name
            self.add_values("length_dist", length_dist)

        if "n_fragments" not in self.anno.columns or pd.isnull(self.anno.loc[sample_name, "n_fragments"]):
            self.add_anno("n_fragments", sample_name, len(frags.frags))
        
        if "chrom_lengths" not in self.uns.keys:
            self.uns["chrom_lengths"] = frags.genome.chrom_sizes
            self.uns["chroms"] = frags.chroms

        if "chroms" not in self.uns.keys:
            self.uns["chroms"] = frags.chroms

        return None



from typing import Dict, List, Tuple, Any
import pandas as pd
import numpy as np
import os
from intervalframe import IntervalFrame
from ngsfragments import Fragments
from ailist import LabeledIntervalArray

# Local imports
from ...core.core import cfDNA
from .wps import wps, coverage


def gene_sum(scores: Dict[str, pd.Series],
             gene_windows: IntervalFrame) -> pd.Series:
    """
    Sum the score values in a window

    Parameters
    ----------
        scores : Dict[str, pd.Series]
            Scores for each chromosome
        gene_windows : IntervalFrame
            Gene associated intervals

    Returns
    -------
        total_sums : pd.Series
            Total sums
    """

    # Initialize results
    chroms = list(scores.keys())
    genes = pd.unique(gene_windows.loc[chroms,"Gene"].values)
    total_sums = pd.Series(np.zeros(len(genes)),
                           index=genes)

    # Iterate over chromosomes
    for chrom in scores:
        chrom_intervals = gene_windows.loc[chrom,:]
        sum_scores = pd.Series(np.zeros(chrom_intervals.shape[0]),
                               index=chrom_intervals.df.loc[:,"Gene"].values)
        
        # Iterate over intervals
        for i, interval in enumerate(chrom_intervals.index):
            sum_scores.values[i] = np.sum(scores[chrom].loc[interval.start:interval.end].values)

        # Sum sums
        summarized_scores = sum_scores.groupby(by=chrom_intervals.df.loc[:,"Gene"].values).sum()
        total_sums.loc[summarized_scores.index] = summarized_scores.values

    return total_sums


def score_gene_sum(frags: Fragments,
                   gene_windows: IntervalFrame,
                   score_method: str = "wps",
                   adjust_lengths: bool = True,
                   protection: int = 120,
                   min_length: int = 120,
                   max_length: int = 210,
                   normalize: bool = True,
                   norm_method: str = "mean",
                   min_gene_length: int = 0,
                   max_gene_length: int = 100000) -> pd.Series:
    """
    Sum scores in all gene_windows
    """

    # Initialize results
    genes = pd.unique(gene_windows.df.loc[:,"Gene"].values)
    results = pd.Series(np.zeros(len(genes)),
                        index=genes)

    # Iterate over chroms
    chroms = set(frags.frags.unique_labels) & set(gene_windows.index.unique_labels)
    for chrom in chroms:

        # Calculate scores
        if score_method == "wps":
            scores = wps(frags, chrom, protection, min_length, max_length, normalize, norm_method)
        elif score_method == "cov":
            scores = frags.coverage(chrom, min_length=min_length, max_length=max_length)
        else:
            raise ModuleNotFoundError("Provided method is not valid [wps, cov]")

        # Sum scores
        sums = gene_sum(scores, gene_windows)
        results.loc[sums.index] = sums.values

    if adjust_lengths:
        results = length_adjust(results, gene_windows, min_length=min_gene_length, max_length=max_gene_length)

    return results


def exon_count(intervals: LabeledIntervalArray,
               gene_windows: IntervalFrame,
               adjust_lengths: bool = True) -> pd.Series:
    """
    Count n hits in a window
    """

    # Count number of hits in gene_windows
    nhits = intervals.nhits_from_LabeledIntervalArray(gene_windows.index)
    summarized_nhits = pd.Series(nhits).groupby(by=gene_windows.df.loc[:,"Gene"].values).sum()

    # Adjust lengths
    if adjust_lengths:
        summarized_nhits = length_adjust(summarized_nhits, gene_windows)

    return summarized_nhits


def length_adjust(values: pd.Series,
                  gene_windows: IntervalFrame,
                  min_length: int = 0,
                  max_length: int = 1000000) -> pd.Series:
    """
    Divide by gene length

    Parameters
    ----------
        values: pd.Series
            Scores
        gene_windows: IntervalFrame
            Gene windows
        min_length: int = 0
            Minimum gene length to include
        max_length: int = 100000
            Maximum gene length to include

    Returns
    -------
        adjusted : pd.Series
            Adjusted scores
    """

    # Find lengths
    starts = gene_windows.index.extract_starts()
    ends = gene_windows.index.extract_ends()
    l = pd.Series(ends-starts).groupby(by=gene_windows.df.loc[:,"Gene"].values).sum()

    # Adjust by dividing
    adjusted = values / l.loc[values.columns.values]

    # Filter by length
    chosen = np.logical_and(l.loc[values.columns.values] > min_length,
                            l.loc[values.columns.values] < max_length)
    adjusted = adjusted.loc[:,chosen.values]

    return adjusted


def call_wps_sum(cfdna_object: cfDNA,
                 frags: Fragments,
                 windows: List[str] = ["exons", "genes", "tss"],
                 genome: str = "hg19",
                 protection: int = 60,
                 min_length: int = 100,
                 max_length: int = 210,
                 normalize: bool = True,
                 norm_method: str = "mean",
                 adjust_lengths: bool = True,
                 use_coverage: bool = False,
                 params: Dict[str,Any] | None = None) -> None:
    """
    """

    # Determine name
    path = os.path.normpath(frags.sam_file)
    file_name = path.split(os.sep)[-1]

    # Log frags
    cfdna_object.log_fragments(frags)

    # Default parameters
    default_params = {"exons": {"upstream": 5000},
                      "genes": {"upstream": 5000,
                                "gene_type": "protein_coding"},
                      "tss": {"upstream": 1000,
                              "downstream": 1000}}
    if params is None:
        params = default_params

    # Get windows
    if genome == "hg19":
        from hg19genome import get_genes as GenomeCaller
    else:
        raise NotImplementedError("Only hg19 is supported currently")
    gene_windows = {w:GenomeCaller(w, **params[w]) for w in windows}

    # Initialize results
    results = {}
    for w in windows:
        genes = pd.unique(gene_windows[w].df.loc[:,"Gene"].values)
        genes.sort()
        results[w] = pd.DataFrame(np.zeros((1, len(genes))),
                        columns=genes, index=[file_name])


        # Iterate over chroms
    for chrom in frags.chroms:
        if use_coverage:
            scores = coverage(frags, chrom, min_length, max_length, normalize, norm_method)
        else:
            scores = wps(frags, chrom, protection, min_length, max_length, normalize, norm_method)
        # Iterate over windows
        for w in windows:
            # Sum scores
            sums = gene_sum(scores, gene_windows[w])
            results[w].loc[file_name, sums.index] = sums.values

    if adjust_lengths:
        for w in windows:
            results[w] = length_adjust(results[w], gene_windows[w])

    for w in windows:
        cfdna_object.add_obs_values(file_name, w+"_sum_wps", results[w])

    return None


from .plot_plt import *
from .plot_bokeh import *
from . import plot_plt
from . import plot_bokeh

__doc__="""
Plotting API
============
.. currentmodule:: cfdna

Generic
-------
.. autosummary::
   :toctree: .

   plot_plt
   plot_bokeh
   
"""


from ngsfragments import Fragments
from intervalframe import IntervalFrame
from math import e
import numpy as np
import pandas as pd
import glob

# Local imports
from ...core import cfDNA
from ...io.read.read_bam import readBam


def score_gene_body(bins: IntervalFrame,
                    start: int,
                    end: int,
                    chrom: str):
    """
    """

    length = end - start
    nhits = bins.intersect(start, end, chrom).df.loc[:,"counts"].values
    #weight = (1 * (1/length)) + e**-1
    #weight = 1 + e**-1
    weight = (1 + (e**-1)) * (1/length)
    score = np.sum(nhits * weight)

    return score


def score_gene_upstream(bins: IntervalFrame,
                        start: int,
                        end: int,
                        chrom: str,
                        upstream: int = 100000):
    """
    """

    length = end - start
    end = start
    start = max(start - upstream, 0)
    nhits = bins.intersect(start, end, chrom)
    
    if nhits.shape[0] > 0:
        distance = end - nhits.ends()
        #weight = ((e**(-abs(distance) / 5000)) * (1/length)) + (e**-1)
        #weight = (e**(-abs(distance) / 5000)) + (e**-1)
        weight = ((e**(-abs(distance) / 5000)) + (e**-1)) * (1/length)
        score = np.sum(nhits.df.loc[:,"counts"].values * weight)
    else:
        score = 0

    return score


def score_gene_downstream(bins: IntervalFrame,
                          start: int,
                          end: int,
                          chrom: str,
                          downstream: int = 100000):
    """
    """

    length = end - start
    start = end
    end = end + downstream
    nhits = bins.intersect(start, end, chrom)
    
    if nhits.shape[0] > 0:
        distance = end - nhits.starts()
        #weight = ((e**(-abs(distance) / 5000)) * (1/length)) + (e**-1)
        #weight = (e**(-abs(distance) / 5000)) + (e**-1)
        weight = ((e**(-abs(distance) / 5000)) + (e**-1)) * (1/length)
        score = np.sum(nhits.df.loc[:,"counts"].values * weight)
    else:
        score = 0

    return score


def gene_activity(frags: Fragments,
                  genome: str = "hg19",
                  min_length: int = 120,
                  max_length: int = 220):
    """
    """
    # Get windows
    if genome == "hg19":
        import hg19genome as Genome
    else:
        raise NotImplementedError("Only hg19 is supported currently")

    # Get genes
    genes = Genome.get_gene_body(upstream=5000, gene_type="protein_coding")

    # Determine chromosomes
    chromosomes = frags.chroms

    # Iterate over chromosomes
    scores = pd.Series(np.zeros(genes.shape[0]),
                       index = genes.loc[:,"Gene"].values)
    j = 0
    for chrom in chromosomes:
        print(chrom)
        # Get genes on chromosome
        chrom_genes = genes.loc[chrom,:]
        # Determine nhits per bin
        bins = frags.frags.get(chrom).bin_nhits(500, min_length, max_length)

        # Create interval frame
        bins_iframe = IntervalFrame(bins[0])
        bins_iframe.df.loc[:,"counts"] = bins[1]

        for i in range(chrom_genes.shape[0]):
            score = score_gene_body(bins_iframe, chrom_genes.index[i].start, chrom_genes.index[i].end, chrom)
            score += score_gene_upstream(bins_iframe, chrom_genes.index[i].start, chrom_genes.index[i].end, chrom)
            score += score_gene_downstream(bins_iframe, chrom_genes.index[i].start, chrom_genes.index[i].end, chrom)
            scores[j] = score
            j += 1

    return scores


def multi_gene_activity(directory: str,
                        genome: str = "hg19",
                        min_length: int = 121,
                        max_length: int = 375,
                        nthreads: int = 1):
    """
    """

    files = glob.glob(directory+"*.bam")

    print(files[0])
    frags = readBam(files[0], nthreads=nthreads)
    scores = gene_activity(frags, genome=genome, min_length=min_length, max_length=max_length)
    scores = scores.to_frame().T
    scores.index = [files[0]]
    for bam in files[1:]:
        print(bam)
        frags = readBam(bam, nthreads=nthreads)
        scores.loc[bam,:] = gene_activity(frags, genome=genome, min_length=min_length, max_length=max_length)

    return scores


from ..cnv.segmentation import call_cnvs
from ..fragmentation.frag_pattern import fragment_profile
from ..nucleosome.nfr import summarize_nfr
from ...core.core import cfDNA
from ngsfragments import fragments
import os
from typing import Dict, List, Tuple


def summarize(cfdna_object: cfDNA,
              frags: fragments,
              cnv_binsize: int = 100000,
              hmm_binsize: int = 1000000,
              method: str = "bcp_online_both",
              outlier_smooth: bool = True,
              gauss_smooth: bool = False,
              bcp_cutoff: float = 0.3,
              nfr_filenames: List[str] = ["hg19_TSS.bed.gz","hg19_CTCF.bed.gz","hg19_TFBS.bed.gz"],
              stranded: List[bool] = [True,False,False],
              normal: List[float] = [0.1, 0.5, 0.9],
              ploidy: List[int] = [2],
              estimatePloidy: bool = False,
              minSegmentBins: int = 25,
              maxCN: int = 7,
              progress_bar: bool = False,
              verbose: bool = False) -> None:
    """
    Calculate summary metrics

    Parameters
    ----------
        bin_size
            int
        genome
            str
        bin_bias_h5_fn
            str
        outlier_smooth
            bool
        gauss_smooth
            bool
        n_threads
            int
        tss_fn
            str
        normal
            list
        ploidy
            list
        estimatePloidy
            bool
        minSegmentBins
            int
        maxCN
            int
        bcp_cutoff
            float

    Returns
    -------
        None: None
            None
    """

    # Determine file name
    path = os.path.normpath(frags.sam_file)
    file_name = path.split(os.sep)[-1]

    # Check if file was previously annotated
    if file_name not in cfdna_object.filenames:
        cfdna_object.log_filename(frags)

    # Call CNVs
    if verbose: print("Running CNV pipeline...", flush=True)
    cfdna_object = call_cnvs(cfdna_object, frags, cnv_binsize=cnv_binsize, hmm_binsize=hmm_binsize, method=method, outlier_smooth=outlier_smooth,
                            gauss_smooth=gauss_smooth, bcp_cutoff=bcp_cutoff, normal=normal, ploidy=ploidy, estimatePloidy=estimatePloidy,
                            minSegmentBins=minSegmentBins, maxCN=maxCN, verbose=verbose)

    # Determine fragmentation profile
    if verbose: print("Calculating fragment profile...", flush=True)
    fragment_profile(cfdna_object, frags, bin_size=cnv_binsize, bin_bias_h5_fn=None,
                     bias_correct=True, smooth=True, verbose=verbose)
    
    # Calculate NFR
    if verbose: print("Calling NFR for long fragments...", flush=True)
    summarize_nfr(cfdna_object, frags, "long", nfr_filenames=nfr_filenames, stranded=stranded,
                  protection=120, min_length=120, max_length=220,
                  cnv_binsize=cnv_binsize, method="mean", peak_min_length=50,
                  peak_max_length=250, progress_bar=progress_bar, verbose=verbose)

    if verbose: print("Calling NFR for short fragments...", flush=True)
    summarize_nfr(cfdna_object, frags, "short", nfr_filenames=nfr_filenames, stranded=stranded,
                  protection=16, min_length=16, max_length=120,
                  cnv_binsize=cnv_binsize, method="mean", peak_min_length=50,
                  peak_max_length=200, progress_bar=progress_bar, verbose=verbose)

    return None

    